# -*- coding: utf-8 -*-
# pylint: disable=no-member, invalid-name
""" common function for simplex """

from itertools import combinations
import numpy as np


def sim_conv(simplex, n=3):
    """simplex to any dimension"""
    v = [list(combinations(sim, n)) for sim in simplex]
    # change to (num_of_points x n)
    t = np.sort(np.array(v).reshape(-1, n), axis=1)
    # delete duplicated entries
    t_unique = np.unique(t.view([("", t.dtype)] * n)).view(np.uint32)
    return t_unique


def sim2tri(simplex):
    """convert simplex of high dimension to indices of triangles"""
    return sim_conv(simplex, 3)


def sim2edge(simplex):
    """convert simplex of high dimension to indices of edges"""
    return sim_conv(simplex, 2)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-arguments
# pylint: disable=too-many-instance-attributes, too-many-locals
# pylint: disable=arguments-differ
""" dynamic EIT solver using JAC """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
import scipy.linalg as la

from .base import EitBase


class JAC(EitBase):
    """A sensitivity-based EIT imaging class"""

    def setup(self, p=0.20, lamb=0.001, method="kotre"):
        """
        JAC, Jacobian matrix based reconstruction.

        Parameters
        ----------
        p, lamb: float
            JAC parameters
        method: str
            regularization methods
        """
        # passing imaging parameters
        self.params = {"p": p, "lamb": lamb, "method": method}
        # pre-compute H0 for dynamical imaging
        # H = (J.T*J + R)^(-1) * J.T
        self.H = h_matrix(self.J, p, lamb, method)

    def map(self, dv):
        """return Hv"""
        return -np.dot(self.H, dv.transpose())

    def solve_gs(self, v1, v0):
        """solving by weighted frequency"""
        a = np.dot(v1, v0) / np.dot(v0, v0)
        dv = v1 - a * v0
        ds = -np.dot(self.H, dv.transpose())
        # return average epsilon on element
        return ds

    def jt_solve(self, v1, v0, normalize=True):
        """
        a 'naive' back projection using the transpose of Jac.
        This scheme is the one published by kotre (1989):

        [1] Kotre, C. J. (1989).
            A sensitivity coefficient method for the reconstruction of
            electrical impedance tomograms.
            Clinical Physics and Physiological Measurement,
            10(3), 275–281. doi:10.1088/0143-0815/10/3/008

        The input (dv) and output (ds) is log-normalized.
        """
        if normalize:
            dv = np.log(np.abs(v1) / np.abs(v0)) * np.sign(v0.real)
        else:
            dv = v1 - v0
        # s_r = J^Tv_r
        ds = -np.dot(self.J.conj().T, dv)
        return np.exp(ds) - 1.0

    def gn(
        self,
        v,
        x0=None,
        maxiter=1,
        gtol=1e-4,
        p=None,
        lamb=None,
        lamb_decay=1.0,
        lamb_min=0,
        method="kotre",
        verbose=False,
        vector=False,
    ):
        """
        Gaussian Newton Static Solver
        You can use a different p, lamb other than the default ones in setup

        Parameters
        ----------
        v: NDArray
            boundary measurement
        x0: NDArray, optional
            initial guess
        maxiter: int, optional
            number of maximum iterations
        p, lamb: float
            JAC parameters (can be overridden)
        lamb_decay: float
            decay of lamb0, i.e., lamb0 = lamb0 * lamb_delay of each iteration
        lamb_min: float
            minimal value of lamb
        method: str, optional
            'kotre' or 'lm'
        verbose: bool, optional
            print debug information
        vector: bool, optional
            Use vectorized methods or regular methods, for compatibility.

        Returns
        -------
        sigma: NDArray
            Complex-valued conductivities

        Note
        ----
        Gauss-Newton Iterative solver,
            x1 = x0 - (J^TJ + lamb*R)^(-1) * r0
        where:
            R = diag(J^TJ)**p
            r0 (residual) = real_measure - forward_v
        """
        if x0 is None:
            x0 = self.perm
        if p is None:
            p = self.params["p"]
        if lamb is None:
            lamb = self.params["lamb"]
        if method is None:
            method = self.params["method"]

        # convergence test
        x0_norm = np.linalg.norm(x0)

        for i in range(maxiter):

            # forward solver
            fs = self.fwd.solve_eit(
                self.ex_mat, step=self.step, perm=x0, parser=self.parser, vector=vector
            )
            # Residual
            r0 = v - fs.v
            jac = fs.jac

            # Damped Gaussian-Newton
            h_mat = h_matrix(jac, p, lamb, method)

            # update
            d_k = np.dot(h_mat, r0)
            x0 = x0 - d_k

            # convergence test
            c = np.linalg.norm(d_k) / x0_norm
            if c < gtol:
                break

            if verbose:
                print("iter = %d, lamb = %f, gtol = %f" % (i, lamb, c))

            # update regularization parameter
            # lambda can be given in user defined decreasing lists
            lamb *= lamb_decay
            if lamb < lamb_min:
                lamb = lamb_min

        return x0

    def project(self, ds):
        """project ds using spatial difference filter (deprecated)

        Parameters
        ----------
        ds: NDArray
            delta sigma (conductivities)

        Returns
        -------
        NDArray
        """
        d_mat = sar(self.tri)
        return np.dot(d_mat, ds)


def h_matrix(jac, p, lamb, method="kotre"):
    """
    JAC method of dynamic EIT solver:
        H = (J.T*J + lamb*R)^(-1) * J.T

    Parameters
    ----------
    jac: NDArray
        Jacobian
    p, lamb: float
        regularization parameters
    method: str, optional
        regularization method

    Returns
    -------
    H: NDArray
        pseudo-inverse matrix of JAC
    """
    j_w_j = np.dot(jac.transpose(), jac)
    if method == "kotre":
        # see adler-dai-lionheart-2007
        # p=0   : noise distribute on the boundary ('dgn')
        # p=0.5 : noise distribute on the middle
        # p=1   : noise distribute on the center ('lm')
        r_mat = np.diag(np.diag(j_w_j)) ** p
    elif method == "lm":
        # Marquardt–Levenberg, 'lm' for short
        # or can be called NOSER, DLS
        r_mat = np.diag(np.diag(j_w_j))
    else:
        # Damped Gauss Newton, 'dgn' for short
        r_mat = np.eye(jac.shape[1])

    # build H
    h_mat = np.dot(la.inv(j_w_j + lamb * r_mat), jac.transpose())
    return h_mat


def sar(el2no):
    """
    extract spatial difference matrix on the neighbors of each element
    in 2D fem using triangular mesh.

    Parameters
    ----------
    el2no: NDArray
        triangle structures

    Returns
    -------
    D: NDArray
        SAR matrix
    """
    ne = el2no.shape[0]
    d_mat = np.eye(ne)
    for i in range(ne):
        ei = el2no[i, :]
        #
        i0 = np.argwhere(el2no == ei[0])[:, 0]
        i1 = np.argwhere(el2no == ei[1])[:, 0]
        i2 = np.argwhere(el2no == ei[2])[:, 0]
        idx = np.unique(np.hstack([i0, i1, i2]))
        # build row-i
        for j in idx:
            d_mat[i, j] = -1
        nn = idx.size - 1
        d_mat[i, i] = nn
    return d_mat



"""
Reconstruction algorithms for EIT.

Arithmetic
----------
  - fem: Finite-element method
  - base: EitBase class
  - bp: Back-projection
  - jac: Jacobian matrix based method
  - greit: The GREIT algorithm
  - utils: EIT related helper function
  - interp2d: Spatial interpolation for EIT
"""



# coding: utf-8
# pylint: disable=invalid-name, no-member, arguments-differ
""" dynamic EIT solver using SVD """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np

from .jac import JAC


class SVD(JAC):
    """implementing a sensitivity-based EIT imaging class"""

    def setup(self, n=25, rcond=1e-2, method="svd"):
        """
        SVD, singular value decomposition based reconstruction.

        Parameters
        ----------
        n: int
            largest n eigenvalues to be kept
        rcond: double
            r-condition number of pinv
        method: string
            'svd': SVD truncation,
            'pinv': pseudo inverse
        """
        # correct n_ord
        nm, ne = self.J.shape
        n_ord = np.min([nm, ne, n])

        # passing imaging parameters
        self.params = {"n": n_ord, "rcond": rcond, "method": method}

        # pre-compute H0 for dynamical imaging
        if method == "svd":
            JtJ = np.dot(self.J.T, self.J)

            # using svd
            # U, s, Ut = np.linalg.svd(JtJ)
            # U = U[:, :n_ord]
            # s = s[:n_ord]

            # using eigh (more faster for large, symmetric matrix)
            s, U = np.linalg.eigh(JtJ)
            idx = np.argsort(s)[::-1]
            s = s[idx[:n_ord]]
            U = U[:, idx[:n_ord]]

            # pseudo inverse
            JtJ_inv = np.dot(U, np.dot(np.diag(s**-1), U.T))
            self.H = np.dot(JtJ_inv, self.J.T)
        elif method == "pinv":
            self.H = np.linalg.pinv(self.J, rcond=rcond)



# coding: utf-8
# pylint: disable=invalid-name, no-member, no-name-in-module
# pylint: disable=too-many-arguments, too-many-locals
# pylint: disable=too-many-instance-attributes
""" implement 2D/3D distmesh """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

from itertools import combinations
import numpy as np
from numpy import sqrt
from scipy.spatial import Delaunay
from scipy.sparse import csr_matrix

from .utils import dist, edge_project

from .shape import thorax


class DISTMESH:
    """class for distmesh"""

    def __init__(
        self,
        fd,
        fh,
        h0=0.1,
        p_fix=None,
        bbox=None,
        density_ctrl_freq=30,
        deltat=0.1,
        dptol=0.001,
        ttol=0.1,
        Fscale=1.2,
        verbose=False,
    ):
        """initial distmesh class

        Parameters
        ----------
        fd : str
            function handle for distance of boundary
        fh : str
            function handle for distance distributions
        h0 : float, optional
            Distance between points in the initial distribution p0,
            default=0.1 For uniform meshes, h(x,y) = constant,
            the element size in the final mesh will usually be
            a little larger than this input.
        p_fix : array_like, optional
            fixed points, default=[]
        bbox : array_like, optional
            bounding box for region, bbox=[xmin, ymin, xmax, ymax].
            default=[-1, -1, 1, 1]
        density_ctrl_freq : int, optional
            cycles of iterations of density control, default=20
        deltat : float, optional
            mapping forces to distances, default=0.2
        dptol : float, optional
            exit criterion for minimal distance all points moved, default=0.01
        ttol : float, optional
            enter criterion for re-delaunay the lattices, default=0.1
        Fscale : float, optional
            rescaled string forces, default=1.2
            if set too small, points near boundary will be pushed back
            if set too large, points will be pushed towards boundary

        Notes
        -----
        """
        # shape description
        self.fd = fd
        self.fh = fh
        self.h0 = h0

        # a small gap, allow points who are slightly outside of the region
        self.deps = np.sqrt(np.finfo(np.double).eps) * h0
        self.geps = 1e-1 * h0

        # control the distmesh computation flow
        self.densityctrlfreq = density_ctrl_freq
        self.dptol = dptol
        self.ttol = ttol
        self.Fscale = Fscale
        self.deltat = deltat

        # default bbox is 2D
        if bbox is None:
            bbox = [[-1, -1], [1, 1]]
        # p : coordinates (x,y) or (x,y,z) of meshes
        self.n_dim = np.shape(bbox)[1]
        if self.n_dim == 2:
            p = bbox2d_init(h0, bbox)
        else:
            p = bbox3d_init(h0, bbox)

        # control debug messages
        self.verbose = verbose
        self.num_triangulate = 0
        self.num_density = 0
        self.num_move = 0

        """
        keep points that are inside the thorax shape using a function that returns a matrix containing
        True if the corresponing point is inside the shape, False if not.
        """
        if fd == thorax:
            p = p[fd(p)]
        else:
            # keep points inside (minus distance) with a small gap (geps)
            p = p[fd(p) < self.geps]  # pylint: disable=E1136

        # rejection points by sampling on fh
        r0 = 1.0 / fh(p) ** self.n_dim
        selection = np.random.rand(p.shape[0]) < (r0 / np.max(r0))
        p = p[selection]

        # specify fixed points
        if p_fix is None:
            p_fix = []
        self.pfix = p_fix
        self.nfix = len(p_fix)

        # convert boolean array to 2D to be compatible with Delaunay pts paramater (must be 2D)
        if fd == thorax:
            p = np.reshape(p, (-1, 2))

        # remove duplicated points of p and p_fix
        # avoid overlapping of mesh points
        if self.nfix > 0:
            p = remove_duplicate_nodes(p, p_fix, self.geps)
            p = np.vstack([p_fix, p])

        if fd == thorax:
            p = np.reshape(
                p, (-1, 2)
            )  # convert boolean array to 2D to be compatible with Delaunay pts paramater (must be 2D)

        # store p and N
        self.N = p.shape[0]
        self.p = p

        # initialize pold with inf: it will be re-triangulate at start
        self.pold = np.inf * np.ones((self.N, self.n_dim))

        # build edges list for triangle or tetrahedral. i.e., in 2D triangle
        # edge_combinations is [[0, 1], [1, 2], [2, 0]]
        self.edge_combinations = list(combinations(range(self.n_dim + 1), 2))

        # triangulate, generate simplices and bars
        self.triangulate()

    def is_retriangulate(self):
        """test whether re-triangulate is needed"""
        return np.max(dist(self.p - self.pold)) > (self.h0 * self.ttol)

    def triangulate(self):
        """retriangle by delaunay"""
        self.debug("enter triangulate = ", self.num_triangulate)
        self.num_triangulate += 1
        # pnew[:] = pold[:] makes a new copy, not reference
        self.pold = self.p.copy()

        # triangles where the points are arranged counterclockwise
        if self.fd != thorax:
            tri = Delaunay(self.p).simplices
        else:
            tri = Delaunay(
                self.p, qhull_options="QJ"
            ).simplices  # QJ parameter so tuples don't exceed boundary

        pmid = np.mean(self.p[tri], axis=1)

        if self.fd != thorax:
            # keeps only interior points
            t = tri[self.fd(pmid) < -self.geps]
        else:
            # adapting returned triangles matrix with the thorax integrated fd
            tri_pmid = [p[0] for p in self.fd(pmid)]
            tri_pmid = np.array(tri_pmid)
            t = tri[tri_pmid]
        # extract edges (bars)
        bars = t[:, self.edge_combinations].reshape((-1, 2))
        # sort and remove duplicated edges, eg (1,2) and (2,1)
        # note : for all edges, non-duplicated edge is boundary edge
        bars = np.sort(bars, axis=1)
        bars_tuple = bars.view([("", bars.dtype)] * bars.shape[1])

        self.bars = np.unique(bars_tuple).view(bars.dtype).reshape((-1, 2))
        self.t = t

    def bar_length(self):
        """the forces of bars (python is by-default row-wise operation)"""
        # two node of a bar
        bars_a, bars_b = self.p[self.bars[:, 0]], self.p[self.bars[:, 1]]
        barvec = bars_a - bars_b

        # L : length of bars, must be column ndarray (2D)
        L = dist(barvec).reshape((-1, 1))
        # density control on bars
        hbars = self.fh((bars_a + bars_b) / 2.0).reshape((-1, 1))
        # L0 : desired lengths (Fscale matters!)
        L0 = hbars * self.Fscale * sqrt(np.sum(L**2) / np.sum(hbars**2))

        return L, L0, barvec

    def bar_force(self, L, L0, barvec):
        """forces on bars"""
        # abs(forces)
        F = np.maximum(L0 - L, 0)
        # normalized and vectorized forces
        Fvec = F * (barvec / L)
        # now, we get forces and sum them up on nodes
        # using sparse matrix to perform automatic summation
        # rows : left, left, right, right (2D)
        # : left, left, left, right, right, right (3D)
        # cols : x, y, x, y (2D)
        # : x, y, z, x, y, z (3D)
        data = np.hstack([Fvec, -Fvec])
        if self.n_dim == 2:
            rows = self.bars[:, [0, 0, 1, 1]]
            cols = np.dot(np.ones(np.shape(F)), np.array([[0, 1, 0, 1]]))
        else:
            rows = self.bars[:, [0, 0, 0, 1, 1, 1]]
            cols = np.dot(np.ones(np.shape(F)), np.array([[0, 1, 2, 0, 1, 2]]))
        # sum nodes at duplicated locations using sparse matrices
        Ftot = csr_matrix(
            (data.reshape(-1), [rows.reshape(-1), cols.reshape(-1)]),
            shape=(self.N, self.n_dim),
        )
        Ftot = Ftot.toarray()
        # zero out forces at fixed points, as they do not move
        Ftot[0 : len(self.pfix)] = 0

        return Ftot

    def density_control(self, L, L0, dscale=3.0):
        """
        Density control - remove points that are too close
        L0 : Kx1, L : Kx1, bars : Kx2
        bars[L0 > 2*L] only returns bar[:, 0] where L0 > 2L
        """
        self.debug("enter density control = ", self.num_density)
        self.num_density += 1
        # print(self.num_density, self.p.shape)
        # quality control
        ixout = (L0 > dscale * L).ravel()
        ixdel = np.setdiff1d(self.bars[ixout, :].reshape(-1), np.arange(self.nfix))
        self.p = self.p[np.setdiff1d(np.arange(self.N), ixdel)]
        # Nold = N
        self.N = self.p.shape[0]
        self.pold = np.inf * np.ones((self.N, self.n_dim))
        # print('density control ratio : %f' % (float(N)/Nold))

    def move_p(self, Ftot):
        """update p"""
        self.debug("  number of moves = ", self.num_move)
        self.num_move += 1
        # move p along forces
        self.p += self.deltat * Ftot

        # if there is any point ends up outside
        # move it back to the closest point on the boundary
        # using the numerical gradient of distance function
        d = self.fd(self.p)
        ix = d > 0
        if ix.any():
            self.p[ix] = edge_project(self.p[ix], self.fd, self.geps)

        # check whether convergence : no big movements
        delta_move = self.deltat * np.max(dist(Ftot[d < -self.geps]))
        self.debug("  delta_move = ", delta_move)
        score = delta_move < self.dptol * self.h0

        return score

    def debug(self, *args):
        """print debug messages"""
        if self.verbose:
            print(*args)


def bbox2d_init(h0, bbox):
    """
    generate points in 2D bbox (not including the ending point of bbox)

    Parameters
    ----------
    h0 : float
        minimal distance of points
    bbox : array_like
        [[x0, y0],
         [x1, y1]]

    Returns
    -------
    array_like
        points in bbox
    """
    x, y = np.meshgrid(
        np.arange(bbox[0][0], bbox[1][0], h0),
        np.arange(bbox[0][1], bbox[1][1], h0 * sqrt(3) / 2.0),
        indexing="xy",
    )
    # shift even rows of x
    # shift every second row h0/2 to the right, therefore,
    # all points will be a distance h0 from their closest neighbors
    x[1::2, :] += h0 / 2.0
    # p : Nx2 ndarray
    p = np.array([x.ravel(), y.ravel()]).T
    return p


def bbox3d_init(h0, bbox):
    """
    generate nodes in 3D bbox

    See Also
    --------
    bbox2d : converting bbox to 2D points
    """
    xspace = h0
    yspace = h0 * sqrt(3) / 2.0
    zspace = h0 * sqrt(3 / 2.0)
    # build meshgrid with Cartesian indexing
    x, y, z = np.meshgrid(
        np.arange(bbox[0][0], bbox[1][0], xspace),
        np.arange(bbox[0][1], bbox[1][1], yspace),
        np.arange(bbox[0][2], bbox[1][2], zspace),
        indexing="xy",
    )

    # shift every second row of x h0/2 to the right, therefore,
    # all points on xy-plane will have equal distance h0
    # from their closest neighbors
    x[1::2, :, :] += h0 / 2.0
    # shift every second z, where x += h0/2, y += h0/(2*sqrt(3))
    # note : in tetrahedral, the distance to all neighbores of a point
    # is not equilength, aka, equi-tetrahedral can not fill space.
    x[:, :, 1::2] += h0 / 2.0
    y[:, :, 1::2] += h0 / (2.0 * sqrt(3))
    # p : Nx3 ndarray
    p = np.array([x.ravel(), y.ravel(), z.ravel()]).T
    return p


def remove_duplicate_nodes(p, pfix, geps):
    """remove duplicate points in p who are closed to pfix. 3D, ND compatible

    Parameters
    ----------
    p : array_like
        points in 2D, 3D, ND
    pfix : array_like
        points that are fixed (can not be moved in distmesh)
    geps : float, optional (default=0.01*h0)
        minimal distance that two points are assumed to be identical

    Returns
    -------
    array_like
        non-duplicated points
    """
    for row in pfix:
        pdist = dist(p - row)
        # extract non-duplicated row slices
        p = p[pdist > geps]
    return p


def build(
    fd, fh, pfix=None, bbox=None, h0=0.1, densityctrlfreq=10, maxiter=500, verbose=False
):
    """main function for distmesh

    See Also
    --------
    DISTMESH : main class for distmesh

    Parameters
    ----------
    maxiter : int, optional
        maximum iteration numbers, default=1000

    Returns
    -------
    p : array_like
        points on 2D bbox
    t : array_like
        triangles describe the mesh structure

    Notes
    -----
    there are many python or hybrid python + C implementations in github,
    this implementation is merely implemented from scratch
    using PER-OLOF PERSSON's Ph.D thesis and SIAM paper.

    .. [1] P.-O. Persson, G. Strang, "A Simple Mesh Generator in MATLAB".
       SIAM Review, Volume 46 (2), pp. 329-345, June 2004

    Also, the user should be aware that, equal-edged tetrahedron cannot fill
    space without gaps. So, in 3D, you can lower dptol, or limit the maximum
    iteration steps.

    """
    # parsing arguments
    # make sure : g_Fscale < 1.5
    mode_3D = False
    if bbox is None:
        g_dptol, g_ttol, g_Fscale, g_deltat = 0.001, 0.1, 1.2, 0.2
    else:
        # perform error check on bbox
        bbox = np.array(bbox)
        if (bbox.ndim == 1) or (bbox.shape[1] not in [2, 3]):
            raise TypeError("only 2D, 3D are supported, bbox = ", bbox)
        if bbox.shape[0] != 2:
            raise TypeError("please specify lower and upper bound of bbox")
        if bbox.shape[1] == 2:
            # default parameters for 2D
            g_dptol, g_ttol, g_Fscale, g_deltat = 0.001, 0.1, 1.3, 0.2
        else:
            # default parameters for 3D
            # g_dptol, g_ttol, g_Fscale = 0.045, 0.150, 1.125
            mode_3D = True
            g_dptol, g_ttol, g_Fscale, g_deltat = 0.001, 0.1, 1.1, 0.1

    # initialize distmesh
    dm = DISTMESH(
        fd,
        fh,
        h0=h0,
        p_fix=pfix,
        bbox=bbox,
        density_ctrl_freq=densityctrlfreq,
        deltat=g_deltat,
        dptol=g_dptol,
        ttol=g_ttol,
        Fscale=g_Fscale,
        verbose=verbose,
    )

    # now iterate to push to equilibrium
    for i in range(maxiter):
        if dm.is_retriangulate():
            # print("triangulate = %d" % dm.num_triangulate)
            dm.triangulate()

        # calculate bar forces
        L, L0, barvec = dm.bar_length()

        # density control
        if not mode_3D:
            if (i % densityctrlfreq) == 0 and (L0 > 2 * L).any():
                dm.density_control(L, L0)
                # continue to triangulate
                continue

        # calculate bar forces
        Ftot = dm.bar_force(L, L0, barvec)

        if fd != thorax:
            # update p
            converge = dm.move_p(Ftot)
            # the stopping ctriterion (movements interior are small)
            if converge:
                break
        else:  # Thorax mesh is created so far without iteration process (to be updated)
            break

    # at the end of iteration, (p - pold) is small, so we recreate delaunay
    dm.triangulate()

    # you should remove duplicate nodes and triangles
    return dm.p, dm.t



# pylint: disable=no-member, invalid-name, too-many-instance-attributes
"""
load raw EWD into memory
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import os
from struct import unpack
import numpy as np


class EWD:
    """EIT raw/waveform data"""

    def __init__(self, file_name):
        """
        RAW data (.EWD) contains only data
        """
        self.file_name = file_name
        self.file_size = os.path.getsize(file_name)
        # 256 measurement waveforms, 128 points per wave, 2 bytes per point
        self.n_wave = 256
        self.n_point = 128
        self.n_data = self.n_wave * self.n_point
        self.frame_size = self.n_data * 2  # signed short
        self.n_frame = int(self.file_size / self.frame_size)
        self.tot_data = int(self.file_size / 2)

        raw = self.load_raw()
        self.wave = self.demodulate(raw)
        scale = 4096 / 65536 / 29.9 * 1000 / 1250
        self.data = (self.wave[:, :256] + 1j * self.wave[:, 256:]) * scale

    def load_raw(self):
        """load raw data"""
        raw = np.zeros((self.n_frame, self.n_data), dtype=np.int)
        with open(self.file_name, "rb") as fh:
            for i in range(self.n_frame):
                d = fh.read(self.frame_size)
                raw[i] = unpack("{}h".format(self.n_data), d)

        return raw

    def demodulate(self, raw):
        """demodulate raw data into [re, im]"""
        wave = np.zeros((self.n_frame, 2 * self.n_wave), dtype=np.double)
        sin_rom = np.sin(2.0 * np.pi * np.arange(self.n_point) / self.n_point)
        cos_rom = np.cos(2.0 * np.pi * np.arange(self.n_point) / self.n_point)
        for i in range(self.n_frame):
            dw = raw[i].reshape(self.n_wave, -1)
            wave_re = np.sum(dw * sin_rom, axis=1)
            wave_im = np.sum(dw * cos_rom, axis=1)
            wave[i] = np.concatenate([wave_re, wave_im]) / (self.n_point / 2.0)

        return wave

    def to_erd(self, src, dst):
        """combine ERD and demodulate EWD to a new file"""
        file_size = os.path.getsize(src)
        header_size = 1024
        data_size = 4096  # 2*256 doubles
        frame_size = header_size + data_size
        n_frame = int(file_size / frame_size)
        assert n_frame == self.n_frame
        with open(src, "rb") as fr, open(dst, "wb") as fw:
            for i in range(n_frame):
                h = fr.read(header_size)
                fr.read(data_size)  # skip data
                d = self.wave[i].tobytes()
                fw.write(h + d)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals, too-many-arguments
""" create multi-shell mesh """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np

from .shape import circle, fix_points_circle
from .distmesh import build
from .utils import check_order
from .mesh_circle import MeshCircle


def multi_shell(n_fan=8, n_layer=8, n_el=16, r_layer=None, perm_per_layer=None):
    """
    create simple multi shell mesh

    Parameters
    ----------
    n_fan : int
        number of fans per layer
    n_layer : int
        number of layers
    n_el : int
        number of electrodes
    r_layer : NDArray
        int, anomaly layers
    perm_per_layer : NDArray
        float, conductivity on each anomaly layer

    Notes
    -----
    The quality of meshes near the boundary is bad.
    (sharp angles, angle of 90, etc.)
    """
    if np.size(r_layer) != np.size(perm_per_layer):
        raise ValueError("r_layer and perm_per_layer must have same length")

    model = MeshCircle(n_fan=n_fan, n_layer=n_layer, n_el=n_el)
    p, e, el_pos = model.create()

    # tweak permittivity
    delta_r = 1.0 / n_layer
    perm = np.ones(e.shape[0])

    t_center = np.mean(p[e], axis=1)
    r_center = np.sqrt(np.sum(t_center**2, axis=1))
    for layer, a in zip(r_layer, perm_per_layer):
        r0, r1 = delta_r * (layer - 1), delta_r * layer
        idx = (r0 < r_center) & (r_center < r1)
        perm[idx] = a

    # 5. build output structure
    mesh = {"element": e, "node": p, "perm": perm}

    return mesh, el_pos


def multi_circle(
    r=1.0, background=1.0, n_el=16, h0=0.006, r_layer=None, perm_per_layer=None, ppl=64
):
    """
    create multi layer circle mesh

    Parameters
    ----------
    r : float
        radius of the circle
    background : float
        background conductivity
    n_el : int
        number of electrodes
    h0 : float
        initial area of meshes
    r_layer : NDArray
        n x p arrays, each row represents [r1, ..., rp] where r1 < r < rp
    perm_per_layer : NDArray
        n x 1 arrays, the conductivity on each layer
    ppl : int
        point per layer

    Notes
    -----
    Due to the size constraints the triangle mesh, layer may be discontinuous
    especially in the interior. However, the quality of meshes is superior
    to multi_shell.
    """

    if np.ndim(perm_per_layer) != 1:
        raise ValueError("perm_per_layer must be 1-dimension")

    if np.shape(r_layer)[0] != np.size(perm_per_layer):
        raise ValueError("r_layer and perm_per_layer must have same length")

    def _fd(pts):
        """shape function"""
        return circle(pts, pc=[0, 0], r=r)

    def _fh(pts):
        """distance function"""
        r2 = np.sum(pts**2, axis=1)
        return 0.6 * (2.0 - r2)

    # 1. build fix points, may be used as the position for electrodes
    if ppl > n_el:
        step = np.ceil(ppl / n_el).astype("int")
        p_fix = fix_points_circle(ppl=step * n_el)
        # generate electrodes, the same as p_fix (top n_el)
        el_pos = np.arange(n_el) * step
    else:
        p_fix = fix_points_circle(n_el)
        el_pos = np.arange(n_el)

    # 2. append fix points on layers
    for layer in r_layer:
        for (i, ri) in enumerate(layer):
            p_fix_layer = ri * r * fix_points_circle(offset=i / 2.0, ppl=ppl)
            p_fix = np.vstack([p_fix, p_fix_layer])

    # 3. build triangle (more frequently control the nodes)
    p, t = build(_fd, _fh, pfix=p_fix, h0=h0, densityctrlfreq=10)

    # check whether t is counter-clock-wise, otherwise reshape it
    t = check_order(p, t)

    # 4. init uniform element sigma
    perm = background * np.ones(t.shape[0])
    t_center = np.mean(p[t], axis=1)
    r_center = np.sqrt(np.sum(t_center**2, axis=1))

    # update permittivity
    for (layer, a) in zip(r_layer, perm_per_layer):
        r0, r1 = np.min(layer), np.max(layer)
        idx = (r0 < r_center) & (r_center < r1)
        perm[idx] = a

    # 5. build output structure
    mesh = {"element": t, "node": p, "perm": perm}

    return mesh, el_pos



# coding: utf-8
# pylint: disable=invalid-name, no-member, no-name-in-module
""" post process for distmesh 2d and 3d """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
import scipy.linalg as la


def dist(p):
    """distances to origin of nodes. '3D', 'ND' compatible

    Parameters
    ----------
    p : array_like
        points in 2D, 3D. i.e., in 3D
        [[x, y, z],
         [2, 3, 3],
         ...
         [1, 2, 1]]

    Returns
    -------
    array_like
        distances of points to origin
    """
    if p.ndim == 1:
        d = np.sqrt(np.sum(p**2))
    else:
        d = np.sqrt(np.sum(p**2, axis=1))

    return d


def edge_project(pts, fd, h0=1.0):
    """project points back on edge"""
    g_vec = edge_grad(pts, fd, h0)
    return pts - g_vec


def edge_grad(p, fd, h0=1.0):
    """
    project points back on the boundary (where fd=0) using numerical gradient
    3D, ND compatible

    Parameters
    ----------
    p : array_like
        points on 2D, 3D
    fd : str
        function handler of distances
    h0 : float
        minimal distance

    Returns
    -------
    array_like
        gradients of points on the boundary

    Note
    ----
        numerical gradient:
        f'_x = (f(p+delta_x) - f(p)) / delta_x
        f'_y = (f(p+delta_y) - f(p)) / delta_y
        f'_z = (f(p+delta_z) - f(p)) / delta_z

        you should specify h0 according to your actual mesh size
    """
    # d_eps = np.sqrt(np.finfo(float).eps)*h0
    # d_eps = np.sqrt(np.finfo(float).eps)
    d_eps = 1e-8 * h0

    # get dimensions
    if np.ndim(p) == 1:
        p = p[:, np.newaxis]

    # distance
    d = fd(p)

    # calculate the gradient of each axis
    ndim = p.shape[1]
    pts_xyz = np.repeat(p, ndim, axis=0)
    delta_xyz = np.repeat([np.eye(ndim)], p.shape[0], axis=0).reshape(-1, ndim)
    deps_xyz = d_eps * delta_xyz
    g_xyz = (fd(pts_xyz + deps_xyz) - np.repeat(d, ndim, axis=0)) / d_eps

    # normalize gradient, avoid divide by zero
    g = g_xyz.reshape(-1, ndim)
    g2 = np.sum(g**2, axis=1)

    # move unit
    g_num = g / g2[:, np.newaxis] * d[:, np.newaxis]

    return g_num


def edge_list(tri):
    """
    edge of delaunay triangles are unique bars, O(n^2)

    besides this duplication test, you may also use fd to test edge bars,
    where the two nodes of any edge both satisfy fd=0 (or with a geps gap)

    Parameters
    ----------
    tri : array_like
        triangles list
    """
    bars = tri[:, [[0, 1], [1, 2], [2, 0]]].reshape((-1, 2))
    bars = np.sort(bars, axis=1)
    bars = bars.view("i, i")
    n = bars.shape[0]

    ix = [True] * n
    for i in range(n - 1):
        # skip if already marked as duplicated
        if ix[i] is not True:
            continue
        # mark duplicate entries, at most 2-duplicate bars, if found, break
        for j in range(i + 1, n):
            if bars[i] == bars[j]:
                ix[i], ix[j] = False, False
                break

    return bars[np.array(ix)].view("i")


def check_order(no2xy, el2no):
    """
    loop over all elements, calculate the Area of Elements (aoe)
    if AOE > 0, then the order of element is correct
    if AOE < 0, reorder the element

    Parameters
    ----------
    no2xy : NDArray
        Nx2 ndarray, (x,y) locations for points
    el2no : NDArray
        Mx3 ndarray, elements (triangles) connectivity

    Returns
    -------
    NDArray
        ae, area of each element

    Notes
    -----
    tetrahedron should be parsed that the sign of volume is [1, -1, 1, -1]
    """
    el_num, n_vertices = np.shape(el2no)
    # select ae function
    if n_vertices == 3:
        _fn = tri_area
    elif n_vertices == 4:
        _fn = tet_volume
    # calculate ae and re-order tri if necessary
    for ei in range(el_num):
        no = el2no[ei, :]
        xy = no2xy[no, :]
        v = _fn(xy)
        # if CCW, area should be positive, otherwise,
        if v < 0:
            el2no[ei, [1, 2]] = el2no[ei, [2, 1]]

    return el2no


def tri_area(xy):
    """
    return area of a triangle, given its tri-coordinates xy

    Parameters
    ----------
    xy : NDArray
        (x,y) of nodes 1,2,3 given in counterclockwise manner

    Returns
    -------
    float
        area of this element
    """
    s = xy[[2, 0]] - xy[[1, 2]]
    a_tot = 0.50 * la.det(s)
    # (should be positive if tri-points are counter-clockwise)
    return a_tot


def tet_volume(xyz):
    """calculate the volume of tetrahedron"""
    s = xyz[[2, 3, 0]] - xyz[[1, 2, 3]]
    v_tot = (1.0 / 6.0) * la.det(s)
    return v_tot


if __name__ == "__main__":
    # test 'edge_project'
    def fd_test(p):
        """unit circle/ball"""
        if len(p.shape) == 1:
            d = np.sqrt(np.sum(p**2)) - 1.0
        else:
            d = np.sqrt(np.sum(p**2, axis=1)) - 1.0

        return d

    p_test = [[1, 2, 3], [2, 2, 2], [1, 3, 3], [1, 1, 1]]
    a = edge_project(p_test, fd_test)

    # test 'edge_list'



# pylint: disable=no-member, invalid-name, no-name-in-module
# pylint: disable=too-many-arguments, too-many-locals
"""using the geometry of mesh to segment the EIT images"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

import numpy as np
from numpy.linalg import eig, inv
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt

from pkg_resources import resource_filename
from pyeit.io import mes


class SimpleMeshGeometry:
    """
    extract segments from meshes using:
    5-13 as the line for left and right hemisphere
    1-9 as the line for up and down

    Returns
    -------
    left, right, upper, down,
    upper-left, upper-right, down-left, down-right
    """

    # constructor
    def __init__(self, mesh, el_pos, method="element"):
        """
        Parameters
        ----------
        mesh : dictionary
            'element', 'node', 'perm'
        el_pos : array
            indices of electrode nodes
        method : string
            'element', 'node'
        """
        if method not in ["element", "node"]:
            raise TypeError("method do not recognized.")
        if method == "element":
            # find the center of elements
            self.ts = self._tri_centers(mesh)
        else:
            self.ts = mesh["node"]

        # vertical cut (from 5->13)
        pts = mesh["node"]
        vert = np.array([pts[el_pos[12]], pts[el_pos[4]]])
        self.vert_vec = vert[1] - vert[0]

        # horizonal cut (from 1->9)
        horz = np.array([pts[el_pos[0]], pts[el_pos[8]]])
        self.horz_vec = horz[1] - horz[0]

        # points vectors
        self.pv_vec = self.ts - vert[0]
        self.pv_vec = self.pv_vec.transpose()
        self.ph_vec = self.ts - horz[0]
        self.ph_vec = self.ph_vec.transpose()

    def left(self):
        """
        extract left coordinates
        note: bmp's right
        """
        return self._line_side(self.vert_vec, self.pv_vec)

    def right(self):
        """
        extract right coordinates
        note : bmp's left
        """
        return np.logical_not(self.left())

    def upper(self):
        """
        extract top coordinates
        note : bmp's bottom
        """
        return self._line_side(self.horz_vec, self.ph_vec)

    def down(self):
        """
        extract bottom coordinates
        note : bmp's top
        """
        return np.logical_not(self.upper())

    def upper_left(self):
        """upper left coordinates"""
        return np.logical_and(self.left(), self.upper())

    def upper_right(self):
        """upper right coordinates"""
        return np.logical_and(self.right(), self.upper())

    def down_left(self):
        """down left coordinates"""
        return np.logical_and(self.left(), self.down())

    def down_right(self):
        """down right coordinates"""
        return np.logical_and(self.right(), self.down())

    @staticmethod
    def _tri_centers(mesh):
        """
        calculate centers (x,y) of each triangles
        """
        el2no = mesh["element"]
        no2xy = mesh["node"]
        ts = np.mean(no2xy[el2no], axis=1)
        return ts

    @staticmethod
    def _line_side(vline, vpoint):
        """
        Parameters
        ----------
        vline : array
            2x1 array, line[0] is the start point while line [1] is the end
            vline is line[1] - line[0]
        vpoint : array
            2x1 [x, y] coordinates, vpoint is point - line[0]

        find whether a pts is on the left of a line
        """
        proj = vline[0] * vpoint[1] - vline[1] * vpoint[0]
        return proj >= 0


class FitEllipse:
    """
    find the enclosing ellipse of a set of points

    see:
    [1] http://nicky.vanforeest.com/misc/fitEllipse/fitEllipse.html
    [2] https://www.cs.cornell.edu/cv/OtherPdf/Ellipse.pdf
    """

    def __init__(self):
        """initialize mesh"""
        self.a = np.zeros(6)
        # self.pts = mesh['node']
        # self.tri = mesh['element']
        # self.hull_points = self.convex_hull_points(self.pts)

    def fit(self, pts):
        """
        find the enclosing ellipse of data points
        """
        hpts = self.convex_hull_points(pts)
        x, y = hpts[:, 0], hpts[:, 1]
        a = self.fit_ellipse(x, y)
        # x_cent, y_cent
        cx, cy = self.ellipse_center(a)
        # semi major, semi minor
        ca, cb = self.ellipse_axis_length(a)
        # offset angle (radius)
        cangle = self.ellipse_angle_of_rotation(a)
        return np.array([cx, cy, ca, cb, cangle])

    @staticmethod
    def convex_hull_points(pts):
        """
        finding the convex hull (points) of data
        """
        cv = ConvexHull(pts)
        hull_nodes = cv.vertices
        return pts[hull_nodes, :]

    @staticmethod
    def fit_ellipse(x, y):
        """
        fit ellipse using least squares
        """
        x = x[:, np.newaxis]
        y = y[:, np.newaxis]
        D = np.hstack((x * x, x * y, y * y, x, y, np.ones_like(x)))
        S = np.dot(D.T, D)
        C = np.zeros([6, 6])
        C[0, 2] = C[2, 0] = 2
        C[1, 1] = -1
        E, V = eig(np.dot(inv(S), C))
        n = np.argmax(np.abs(E))
        a = V[:, n]
        return a

    @staticmethod
    def ellipse_center(a):
        """
        Returns
        -------
        x_center, y_center
        """
        b, c, d, f, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[0]
        num = b * b - a * c
        x0 = (c * d - b * f) / num
        y0 = (a * f - b * d) / num
        return np.array([x0, y0])

    @staticmethod
    def ellipse_axis_length(a):
        """
        Returns
        -------
        semiminor, semimajor axis length
        """
        b, c, d, f, g, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5], a[0]
        up = 2 * (a * f * f + c * d * d + g * b * b - 2 * b * d * f - a * c * g)
        down1 = (b * b - a * c) * (
            (c - a) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a)
        )
        down2 = (b * b - a * c) * (
            (a - c) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a)
        )
        res1 = np.sqrt(up / down1)
        res2 = np.sqrt(up / down2)
        return np.array([res1, res2])

    @staticmethod
    def ellipse_angle_of_rotation(a):
        """
        Returns
        -------
        ellipse angle in radius
        """
        b, c, a = a[1] / 2, a[2], a[0]
        return 0.5 * np.arctan(2 * b / (a - c))

    @staticmethod
    def ellipse_angle_of_rotation2(a):
        """
        see:
        [1] http://mathworld.wolfram.com/Ellipse.html
        """
        b, c, a = a[1] / 2, a[2], a[0]
        if b == 0:
            if a > c:
                phi = 0
            else:
                phi = np.pi / 2
        else:
            if a > c:
                phi = np.arctan(2 * b / (a - c)) / 2
            else:
                phi = np.pi / 2 + np.arctan(2 * b / (a - c)) / 2

        return phi


def ellipse_points(x_cent=0, y_cent=0, semimaj=1, semimin=1, phi=0, theta_num=1e3):
    """
    see:
    https://casper.berkeley.edu/astrobaki/index.php/Plotting_Ellipses_in_Python

    Create ellipse points.
    The function creates a 2D ellipse in polar coordinates
    then transforms to cartesian coordinates.
    (un-implemented)
    It can take a covariance matrix and plot contours from it.

    x_cent : float
        X coordinate center

    y_cent : float
        Y coordinate center

    semimaj : float
        length of semimajor axis
        always taken to be some phi (-90<phi<90 deg) from
        positive x-axis!

    semimin : float
        length of semiminor axis

    phi : float
        angle in radians of semimajor axis above positive x axis

    theta_num : int
        Number of points to sample along ellipse from 0-2pi

    """
    # Generate data for ellipse structure
    theta = np.linspace(0, 2 * np.pi, int(theta_num))
    r = 1 / np.sqrt((np.cos(theta)) ** 2 + (np.sin(theta)) ** 2)
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    data = np.array([x, y])
    S = np.array([[semimaj, 0], [0, semimin]])
    R = np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])
    T = np.dot(R, S)
    data = np.dot(T, data)
    data[0] += x_cent
    data[1] += y_cent

    return data.transpose()


if __name__ == "__main__":
    # load package mesh data for pyEIT (data/*.mes)
    mstr = resource_filename("eitmesh", "data/DLS2.mes")
    print(mstr)

    # load mesh
    mesh, el_pos = mes.load(mstr)
    pts = mesh["node"]
    tri = mesh["element"]
    x, y = pts[:, 0], pts[:, 1]

    # 1. demo using ellipse fit
    # load shape
    eobj = FitEllipse()
    hpts = eobj.convex_hull_points(pts)
    cvx = hpts[:, 0]
    cvy = hpts[:, 1]

    # ellipse fit
    v = eobj.fit(pts)
    x_cent, y_cent, semimaj, semimin, phi = v
    epts = ellipse_points(x_cent, y_cent, semimaj, semimin, phi)

    # plot mesh
    fig, ax = plt.subplots(figsize=(9, 6))
    mesh_image = mstr.replace(".mes", ".bmp")
    im = plt.imread(mesh_image)
    ax.imshow(im)
    ax.triplot(x, y, tri, "g-", lw=1.0, alpha=0.5)
    ax.set_aspect("equal")
    # plot convex hulls of mesh
    ax.plot(cvx, cvy, "go")
    ax.plot(epts[:, 0], epts[:, 1], "b-", lw=2.0)
    # plot electrodes and its numbering
    ax.plot(pts[el_pos, 0], pts[el_pos, 1], "bo")
    for i, e in enumerate(el_pos):
        ax.text(pts[e, 0], pts[e, 1], str(i + 1), color="r")

    # 2. demo using simple fit
    mg = SimpleMeshGeometry(mesh, el_pos)
    perm = np.zeros(tri.shape[0])
    fig, ax = plt.subplots(figsize=(9, 6))
    mesh_image = mstr.replace(".mes", ".bmp")
    im = plt.imread(mesh_image)
    ax.imshow(im)
    # plot regions
    perm[mg.left()] = 1.0
    perm[mg.down_right()] = 2.0
    img = ax.tripcolor(pts[:, 0], pts[:, 1], tri, perm, shading="flat", alpha=0.50)
    # plot electrodes and its numbering
    ax.plot(pts[el_pos, 0], pts[el_pos, 1], "bo")
    ax.set_title("left = 1.0, down right = 2.0")
    for i, e in enumerate(el_pos):
        ax.text(pts[e, 0], pts[e, 1], str(i + 1), color="r", size=12)
    fig.colorbar(img)

    plt.show()



# copied from https://github.com/adrienverge/yamllint/
import pathlib
from setuptools import setup

from pyeit import (
    __author__,
    __author_email__,
    __license__,
    APP_NAME,
    APP_VERSION,
    APP_DESCRIPTION,
)


HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
setup(
    name=APP_NAME,
    version=APP_VERSION,
    author=__author__,
    author_email=__author_email__,
    description=APP_DESCRIPTION.split("\n")[0],
    long_description=README,
    long_description_content_type="text/markdown",
    license=__license__,
)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-arguments
""" wrapper function of distmesh for EIT """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np

from .distmesh import build
from .mesh_circle import MeshCircle
from .utils import check_order
from .shape import circle, area_uniform, ball, thorax, L_shaped
from .shape import fix_points_fd, fix_points_ball


def create(n_el=16, fd=None, fh=area_uniform, h0=0.1, p_fix=None, bbox=None):
    """
    Generating 2D/3D meshes using distmesh (pyEIT built-in)

    Parameters
    ----------
    n_el: int
        number of electrodes (point-type electrode)
    fd: function
        distance function (circle in 2D, ball in 3D)
    fh: function
        mesh size quality control function
    p_fix: NDArray
        fixed points
    bbox: NDArray
        bounding box
    h0: float
        initial mesh size, default=0.1

    Returns
    -------
    mesh_obj: dict
        {'element', 'node', 'perm'}
    """

    # test conditions if fd or/and bbox are none

    if bbox is None:
        if fd != ball:
            bbox = np.array([[-1, -1], [1, 1]])
        else:
            bbox = [[-1.2, -1.2, -1.2], [1.2, 1.2, 1.2]]

    bbox = np.array(
        bbox
    )  # list is converted to Numpy array so we can use it then (calling shape method..)
    n_dim = bbox.shape[1]  # bring dimension

    # infer dim
    if fd is None:
        if n_dim == 2:
            fd = circle
        elif n_dim == 3:
            fd = ball

    if n_dim not in [2, 3]:
        raise TypeError("distmesh only supports 2D or 3D")
    if bbox.shape[0] != 2:
        raise TypeError("please specify lower and upper bound of bbox")

    if p_fix is None:
        if n_dim == 2:
            if fd == thorax:
                # thorax shape is generated so far without fixed points (to be updated later)
                p_fix = [
                    (-0.098, -0.6463),
                    (-0.4181, -0.6074),
                    (-0.7207, -0.4946),
                    (-0.933, -0.2647),
                    (-0.9147, 0.0543),
                    (-0.8022, 0.3565),
                    (-0.5791, 0.5864),
                    (-0.1653, 0.6819),
                    (0.1564, 0.6571),
                    (0.5814, 0.6353),
                    (0.8298, 0.433),
                    (0.9698, 0.1431),
                    (0.9914, -0.1767),
                    (0.8359, -0.449),
                    (0.5419, -0.5833),
                    (0.2243, -0.6456),
                ]
                p_fix = np.array(p_fix)
            elif fd == L_shaped:
                p_fix = [
                    [1, 0],
                    [1, -1],
                    [0, -1],
                    [-1, -1],
                    [-1, 0],
                    [-1, 1],
                    [0, 1],
                    [0, 0],
                ]  # values brought from distmesh2D L shaped mesh example
                p_fix = np.array(p_fix)
                h0 = 0.15
            else:
                p_fix = fix_points_fd(fd, n_el=n_el)
        elif n_dim == 3:
            p_fix = fix_points_ball(n_el=n_el)

    # 1. build mesh
    p, t = build(fd, fh, pfix=p_fix, bbox=bbox, h0=h0)
    # 2. check whether t is counter-clock-wise, otherwise reshape it
    t = check_order(p, t)
    # 3. generate electrodes, the same as p_fix (top n_el)
    el_pos = np.arange(n_el)
    # 4. init uniform element permittivity (sigma)
    perm = np.ones(t.shape[0], dtype=np.float)
    # 5. build output structure
    mesh = {"element": t, "node": p, "perm": perm}
    return mesh, el_pos


def set_perm(mesh, anomaly=None, background=None):
    """wrapper for pyEIT interface

    Note
    ----
    update permittivity of mesh, if specified.

    Parameters
    ----------
    mesh: dict
        mesh structure
    anomaly: dict, optional
        anomaly is a dictionary (or arrays of dictionary) contains,
        {'x': val, 'y': val, 'd': val, 'perm': val}
        all permittivity on triangles whose distance to (x,y) are less than (d)
        will be replaced with a new value, 'perm' may be a complex value.
    background: float, optional
        set background permittivity

    Returns
    -------
    mesh_obj: dict
        updated mesh structure, {'element', 'node', 'perm'}
    """
    pts = mesh["element"]
    tri = mesh["node"]
    perm = mesh["perm"].copy()
    tri_centers = np.mean(tri[pts], axis=1)

    # this code is equivalent to:
    # >>> N = np.shape(tri)[0]
    # >>> for i in range(N):
    # >>>     tri_centers[i] = np.mean(pts[tri[i]], axis=0)
    # >>> plt.plot(tri_centers[:,0], tri_centers[:,1], 'kx')
    n = np.size(mesh["perm"])

    # reset background if needed
    if background is not None:
        perm = background * np.ones(n)

    # change dtype to 'complex' for complex-valued permittivity
    if anomaly is not None:
        for attr in anomaly:
            if np.iscomplex(attr["perm"]):
                perm = perm.astype("complex")
                break

    # assign anomaly values (for elements in regions)
    if anomaly is not None:
        for _, attr in enumerate(anomaly):
            d = attr["d"]
            # find elements whose distance to (cx,cy) is smaller than d
            if "z" in attr:
                index = (
                    np.sqrt(
                        (tri_centers[:, 0] - attr["x"]) ** 2
                        + (tri_centers[:, 1] - attr["y"]) ** 2
                        + (tri_centers[:, 2] - attr["z"]) ** 2
                    )
                    < d
                )
            else:
                index = (
                    np.sqrt(
                        (tri_centers[:, 0] - attr["x"]) ** 2
                        + (tri_centers[:, 1] - attr["y"]) ** 2
                    )
                    < d
                )
            # update permittivity within indices
            perm[index] = attr["perm"]

    mesh_new = {"node": tri, "element": pts, "perm": perm}
    return mesh_new


def layer_circle(n_el=16, n_fan=8, n_layer=8):
    """generate mesh on unit-circle"""
    model = MeshCircle(n_fan=n_fan, n_layer=n_layer, n_el=n_el)
    p, e, el_pos = model.create()
    perm = np.ones(e.shape[0])

    mesh = {"element": e, "node": p, "perm": perm}
    return mesh, el_pos
